#include "rebocap_pose_node.h"

#include "Roles/LiveLinkAnimationRole.h"
#include "rebocap_skeleton_data.h"

void FRebocapPoseNode::OnInitializeAnimInstance(const FAnimInstanceProxy* InProxy, const UAnimInstance* InAnimInstance) {
  Super::OnInitializeAnimInstance(InProxy, InAnimInstance);
}

void FRebocapPoseNode::GatherDebugData(FNodeDebugData& DebugData) {
  FString DebugLine = DebugData.GetNodeName(this);
  DebugData.AddDebugItem(DebugLine);
  ComponentPose.GatherDebugData(DebugData);
}

static inline void original_transform(FBoneReference bone, EBoneControlSpace space, USkeletalMeshComponent* skel_comp,
                                      FCSPose<FCompactPose>& mesh_bases, std::optional<FTransform>& trans) {
  if (trans) {
    return;
  }
  const FBoneContainer& bone_container = mesh_bases.GetPose().GetBoneContainer();
  FCompactPoseBoneIndex compact_pose_bone_to_modify = bone.GetCompactPoseIndex(bone_container);
  if (compact_pose_bone_to_modify == -1) {
    return;
  }
  FTransform new_bone_tm = mesh_bases.GetComponentSpaceTransform(compact_pose_bone_to_modify);
  FAnimationRuntime::ConvertCSTransformToBoneSpace(skel_comp->GetComponentTransform(), mesh_bases, new_bone_tm, compact_pose_bone_to_modify,
                                                   space);
  trans = new_bone_tm;
}

static inline FQuat get_rotation3(const FName& bone_name, FLiveLinkSubjectFrameData& in_subject_frame_data) {
  FLiveLinkSkeletonStaticData* skeleto_data = in_subject_frame_data.StaticData.Cast<FLiveLinkSkeletonStaticData>();
  FLiveLinkAnimationFrameData* frame_data = in_subject_frame_data.FrameData.Cast<FLiveLinkAnimationFrameData>();
  int32 index = -1;
  if (auto rebocap_bone = skeleto_data->BoneNames.Find(bone_name, index)) {
    return frame_data->Transforms[index].GetRotation();
  }
  return FQuat::Identity;
}

static inline void get_rotation3_location3(const FName& bone_name, FLiveLinkSubjectFrameData& in_subject_frame_data, FQuat& rotation,
                                           FVector& location) {
  FLiveLinkSkeletonStaticData* SkeletonData = in_subject_frame_data.StaticData.Cast<FLiveLinkSkeletonStaticData>();
  FLiveLinkAnimationFrameData* frame_data = in_subject_frame_data.FrameData.Cast<FLiveLinkAnimationFrameData>();
  int32 index = -1;
  if (auto rebocap_bone = SkeletonData->BoneNames.Find(bone_name, index)) {
    rotation = frame_data->Transforms[index].GetRotation();
    location = frame_data->Transforms[index].GetLocation();
    return;
  }
  rotation = FQuat::Identity;
  location = FVector::ZeroVector;
}

static inline void apply_rebocap_transform(FBoneReference bone, FQuat quat, FVector position, EBoneControlSpace space,
                                           USkeletalMeshComponent* skel_comp, FCSPose<FCompactPose>& mesh_bases) {
  if (bone.BoneIndex == -1) {
    return;
  }
  const auto& bone_container = mesh_bases.GetPose().GetBoneContainer();
  FCompactPoseBoneIndex compact_pose_bone_to_modify = bone.GetCompactPoseIndex(bone_container);
  FTransform new_bone_tm = mesh_bases.GetComponentSpaceTransform(compact_pose_bone_to_modify);
  FAnimationRuntime::ConvertCSTransformToBoneSpace(skel_comp->GetComponentTransform(), mesh_bases, new_bone_tm, compact_pose_bone_to_modify,
                                                   space);
  new_bone_tm.SetRotation(quat);
  new_bone_tm.SetTranslation(position);
  FAnimationRuntime::ConvertBoneSpaceTransformToCS(skel_comp->GetComponentTransform(), mesh_bases, new_bone_tm, compact_pose_bone_to_modify,
                                                   space);
  mesh_bases.SetComponentSpaceTransform(compact_pose_bone_to_modify, new_bone_tm);
}

static inline void apply_rebocap_transform(FBoneReference bone, FQuat quat, EBoneControlSpace space, USkeletalMeshComponent* skel_comp,
                                           FCSPose<FCompactPose>& mesh_bases) {
  if (bone.BoneIndex == -1) {
    return;
  }
  const auto& bone_container = mesh_bases.GetPose().GetBoneContainer();
  FCompactPoseBoneIndex compact_pose_bone_to_modify = bone.GetCompactPoseIndex(bone_container);
  FTransform new_bone_tm = mesh_bases.GetComponentSpaceTransform(compact_pose_bone_to_modify);
  FAnimationRuntime::ConvertCSTransformToBoneSpace(skel_comp->GetComponentTransform(), mesh_bases, new_bone_tm, compact_pose_bone_to_modify,
                                                   space);
  new_bone_tm.SetRotation(quat);
  FAnimationRuntime::ConvertBoneSpaceTransformToCS(skel_comp->GetComponentTransform(), mesh_bases, new_bone_tm, compact_pose_bone_to_modify,
                                                   space);
  mesh_bases.SetComponentSpaceTransform(compact_pose_bone_to_modify, new_bone_tm);
}

void FRebocapPoseNode::EvaluateSkeletalControl_AnyThread(FComponentSpacePoseContext& Output, TArray<FBoneTransform>& OutBoneTransforms) {
  USkeletalMeshComponent* skel_comp = Output.AnimInstanceProxy->GetSkelMeshComponent();
  FCSPose<FCompactPose>& mesh_bases = Output.Pose;
  check(OutBoneTransforms.Num() == 0);
  if (!live_link_client_) {
    return;
  }
  FLiveLinkSubjectFrameData subject_frame_data;
  FLiveLinkSubjectName live_link_subject_name = FName("rebocap");
  TSubclassOf<ULiveLinkRole> subject_role = live_link_client_->GetSubjectRole_AnyThread(live_link_subject_name);
  if (!subject_role) {
    return;
  }
  if (subject_role->IsChildOf(ULiveLinkAnimationRole::StaticClass())) {
    if (live_link_client_->EvaluateFrame_AnyThread(live_link_subject_name, ULiveLinkAnimationRole::StaticClass(), subject_frame_data)) {
    } else {
      return;
    }
  }
  FLiveLinkSkeletonStaticData* skeleton_data = subject_frame_data.StaticData.Cast<FLiveLinkSkeletonStaticData>();
  FLiveLinkAnimationFrameData* frame_data = subject_frame_data.FrameData.Cast<FLiveLinkAnimationFrameData>();
  check(skeleton_data);
  check(frame_data);
  EBoneControlSpace test_bone_control_space = BCS_ComponentSpace;
  original_transform(bone_map_.Pelvis, test_bone_control_space, skel_comp, mesh_bases, t_pose_.pelvis);
  original_transform(bone_map_.L_Hip, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_hip);
  original_transform(bone_map_.R_Hip, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_hip);
  original_transform(bone_map_.Spine1, test_bone_control_space, skel_comp, mesh_bases, t_pose_.spine1);
  original_transform(bone_map_.L_Knee, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_knee);
  original_transform(bone_map_.R_Knee, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_knee);
  original_transform(bone_map_.Spine2, test_bone_control_space, skel_comp, mesh_bases, t_pose_.spine2);
  original_transform(bone_map_.L_Ankle, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_ankle);
  original_transform(bone_map_.R_Ankle, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_ankle);
  original_transform(bone_map_.Spine3, test_bone_control_space, skel_comp, mesh_bases, t_pose_.spine3);
  original_transform(bone_map_.L_Foot, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_foot);
  original_transform(bone_map_.R_Foot, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_foot);
  original_transform(bone_map_.Neck, test_bone_control_space, skel_comp, mesh_bases, t_pose_.neck);
  original_transform(bone_map_.L_Collar, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_collar);
  original_transform(bone_map_.R_Collar, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_collar);
  original_transform(bone_map_.Head, test_bone_control_space, skel_comp, mesh_bases, t_pose_.head);
  original_transform(bone_map_.L_Shoulder, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_shoulder);
  original_transform(bone_map_.R_Shoulder, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_shoulder);
  original_transform(bone_map_.L_Elbow, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_elbow);
  original_transform(bone_map_.R_Elbow, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_elbow);
  original_transform(bone_map_.L_Wrist, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_wrist);
  original_transform(bone_map_.R_Wrist, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_wrist);
  original_transform(bone_map_.L_Hand, test_bone_control_space, skel_comp, mesh_bases, t_pose_.l_hand);
  original_transform(bone_map_.R_Hand, test_bone_control_space, skel_comp, mesh_bases, t_pose_.r_hand);
  FQuat pelvis_quat;
  FVector pelvis_position;
  get_rotation3_location3(rebocap_bones::pelvis, subject_frame_data, pelvis_quat, pelvis_position);
  auto l_hip_quat = get_rotation3(rebocap_bones::l_hip, subject_frame_data);
  auto r_hip_quat = get_rotation3(rebocap_bones::r_hip, subject_frame_data);
  auto spine1_quat = get_rotation3(rebocap_bones::spine1, subject_frame_data);
  auto l_knee_quat = get_rotation3(rebocap_bones::l_knee, subject_frame_data);
  auto r_knee_quat = get_rotation3(rebocap_bones::r_knee, subject_frame_data);
  auto spine2_quat = get_rotation3(rebocap_bones::spine2, subject_frame_data);
  auto l_ankle_quat = get_rotation3(rebocap_bones::l_ankle, subject_frame_data);
  auto r_ankle_quat = get_rotation3(rebocap_bones::r_ankle, subject_frame_data);
  auto spine3_quat = get_rotation3(rebocap_bones::spine3, subject_frame_data);
  auto l_foot_quat = get_rotation3(rebocap_bones::l_foot, subject_frame_data);
  auto r_foot_quat = get_rotation3(rebocap_bones::r_foot, subject_frame_data);
  auto neck_quat = get_rotation3(rebocap_bones::neck, subject_frame_data);
  auto l_collar_quat = get_rotation3(rebocap_bones::l_collar, subject_frame_data);
  auto r_collar_quat = get_rotation3(rebocap_bones::r_collar, subject_frame_data);
  auto head_quat = get_rotation3(rebocap_bones::head, subject_frame_data);
  auto l_shoulder_quat = get_rotation3(rebocap_bones::l_shoulder, subject_frame_data);
  auto r_shoulder_quat = get_rotation3(rebocap_bones::r_shoulder, subject_frame_data);
  auto l_elbow_quat = get_rotation3(rebocap_bones::l_elbow, subject_frame_data);
  auto r_elbow_quat = get_rotation3(rebocap_bones::r_elbow, subject_frame_data);
  auto l_wrist_quat = get_rotation3(rebocap_bones::l_wrist, subject_frame_data);
  auto r_wrist_quat = get_rotation3(rebocap_bones::r_wrist, subject_frame_data);
  auto l_hand_quat = get_rotation3(rebocap_bones::l_hand, subject_frame_data);
  auto r_hand_quat = get_rotation3(rebocap_bones::r_hand, subject_frame_data);

  const FBoneContainer& bone_container = mesh_bases.GetPose().GetBoneContainer();
  FCompactPoseBoneIndex compact_pose_bone_to_modify = bone_map_.Pelvis.GetCompactPoseIndex(bone_container);
  if (compact_pose_bone_to_modify == -1) {
    return;
  }
  FTransform new_bone_tm = mesh_bases.GetLocalSpaceTransform(compact_pose_bone_to_modify);
  float testval1 = new_bone_tm.GetLocation().Size();
  float scale_test = testval1 * .01f;

  {
    auto location = (pelvis_position * scale_test);
    apply_rebocap_transform(bone_map_.Pelvis, pelvis_quat * t_pose_.pelvis.value_or(FTransform()).GetRotation(), location,
                            test_bone_control_space, skel_comp, mesh_bases);
  }
#define TRANSFORM(NAME1, NAME2)                                                                                                           \
  apply_rebocap_transform(bone_map_.##NAME1, NAME2##_quat* t_pose_.##NAME2.value_or(FTransform()).GetRotation(), test_bone_control_space, \
                          skel_comp, mesh_bases)
  TRANSFORM(L_Hip, l_hip);
  TRANSFORM(R_Hip, r_hip);
  TRANSFORM(Spine1, spine1);
  TRANSFORM(L_Knee, l_knee);
  TRANSFORM(R_Knee, r_knee);
  TRANSFORM(Spine2, spine2);
  TRANSFORM(L_Ankle, l_ankle);
  TRANSFORM(R_Ankle, r_ankle);
  TRANSFORM(Spine3, spine3);
  TRANSFORM(L_Foot, l_foot);
  TRANSFORM(R_Foot, r_foot);
  TRANSFORM(Neck, neck);
  TRANSFORM(L_Collar, l_collar);
  TRANSFORM(R_Collar, r_collar);
  TRANSFORM(Head, head);
  TRANSFORM(L_Shoulder, l_shoulder);
  TRANSFORM(R_Shoulder, r_shoulder);
  TRANSFORM(L_Elbow, l_elbow);
  TRANSFORM(R_Elbow, r_elbow);
  TRANSFORM(L_Wrist, l_wrist);
  TRANSFORM(R_Wrist, r_wrist);
  TRANSFORM(L_Hand, l_hand);
  TRANSFORM(R_Hand, r_hand);

#undef TRANSFORM
}

void FRebocapPoseNode::PreUpdate(const UAnimInstance* InAnimInstance) {
  Super::PreUpdate(InAnimInstance);
  ILiveLinkClient* this_frame_client = nullptr;
  IModularFeatures& ModularFeatures = IModularFeatures::Get();
  if (ModularFeatures.IsModularFeatureAvailable(ILiveLinkClient::ModularFeatureName)) {
    this_frame_client = &IModularFeatures::Get().GetModularFeature<ILiveLinkClient>(ILiveLinkClient::ModularFeatureName);
  }
  live_link_client_ = this_frame_client;
  {
    UClass* retarget_asset_ptr = retarget_asset_.Get();
    if (!retarget_asset_ptr || retarget_asset_ptr->HasAnyClassFlags(CLASS_Abstract)) {
      retarget_asset_ptr = ULiveLinkRemapAsset::StaticClass();
      retarget_asset_ = retarget_asset_ptr;
    }
    if (!current_retarget_asset_ || retarget_asset_ptr != current_retarget_asset_->GetClass()) {
      current_retarget_asset_ = NewObject<ULiveLinkRemapAsset>(const_cast<UAnimInstance*>(InAnimInstance), retarget_asset_ptr);
      current_retarget_asset_->Initialize();
    }
  }
}

void FRebocapPoseNode::InitializeBoneReferences(const FBoneContainer& RequiredBones) {
  if (current_retarget_asset_ != nullptr && current_retarget_asset_->IsValidLowLevel()) {
    InitBoneMap();
  }
  bone_map_.Pelvis.Initialize(RequiredBones);
  bone_map_.L_Hip.Initialize(RequiredBones);
  bone_map_.R_Hip.Initialize(RequiredBones);
  bone_map_.Spine1.Initialize(RequiredBones);
  bone_map_.L_Knee.Initialize(RequiredBones);
  bone_map_.R_Knee.Initialize(RequiredBones);
  bone_map_.Spine2.Initialize(RequiredBones);
  bone_map_.L_Ankle.Initialize(RequiredBones);
  bone_map_.R_Ankle.Initialize(RequiredBones);
  bone_map_.Spine3.Initialize(RequiredBones);
  bone_map_.L_Foot.Initialize(RequiredBones);
  bone_map_.R_Foot.Initialize(RequiredBones);
  bone_map_.Neck.Initialize(RequiredBones);
  bone_map_.L_Collar.Initialize(RequiredBones);
  bone_map_.R_Collar.Initialize(RequiredBones);
  bone_map_.Head.Initialize(RequiredBones);
  bone_map_.L_Shoulder.Initialize(RequiredBones);
  bone_map_.R_Shoulder.Initialize(RequiredBones);
  bone_map_.L_Elbow.Initialize(RequiredBones);
  bone_map_.R_Elbow.Initialize(RequiredBones);
  bone_map_.L_Wrist.Initialize(RequiredBones);
  bone_map_.R_Wrist.Initialize(RequiredBones);
  bone_map_.L_Hand.Initialize(RequiredBones);
  bone_map_.R_Hand.Initialize(RequiredBones);
}

void FRebocapPoseNode::InitBoneMap() {
  bone_map_.Pelvis = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::pelvis);
  bone_map_.L_Hip = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_hip);
  bone_map_.R_Hip = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_hip);
  bone_map_.Spine1 = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::spine1);
  bone_map_.L_Knee = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_knee);
  bone_map_.R_Knee = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_knee);
  bone_map_.Spine2 = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::spine2);
  bone_map_.L_Ankle = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_ankle);
  bone_map_.R_Ankle = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_ankle);
  bone_map_.Spine3 = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::spine3);
  bone_map_.L_Foot = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_foot);
  bone_map_.R_Foot = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_foot);
  bone_map_.Neck = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::neck);
  bone_map_.L_Collar = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_collar);
  bone_map_.R_Collar = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_collar);
  bone_map_.Head = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::head);
  bone_map_.L_Shoulder = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_shoulder);
  bone_map_.R_Shoulder = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_shoulder);
  bone_map_.L_Elbow = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_elbow);
  bone_map_.R_Elbow = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_elbow);
  bone_map_.L_Wrist = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_wrist);
  bone_map_.R_Wrist = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_wrist);
  bone_map_.L_Hand = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::l_hand);
  bone_map_.R_Hand = current_retarget_asset_->GetRemappedBoneName(rebocap_bones::r_hand);
}

void FRebocapPoseNode::UpdateComponentPose_AnyThread(const FAnimationUpdateContext& Context) {
  GetEvaluateGraphExposedInputs().Execute(Context);
  Super::UpdateComponentPose_AnyThread(Context);
}

void FRebocapPoseNode::UpdateInternal(const FAnimationUpdateContext& Context) {
  GetEvaluateGraphExposedInputs().Execute(Context);
  Super::UpdateInternal(Context);
}

void FRebocapPoseNode::Initialize_AnyThread(const FAnimationInitializeContext& Context) {
  GetEvaluateGraphExposedInputs().Execute(Context);
  Super::Initialize_AnyThread(Context);
}